﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

using System.IO;
using System.Diagnostics;
using System.Threading;

namespace AsyncProcessTable_WinForms {
	public partial class AppForm : Form {
		private static TaskScheduler uiTaskScheduler = null;
		private static AppForm instance = null;

		public AppForm() {
			Debug.Assert(instance == null);

			instance = this;
			uiTaskScheduler = TaskScheduler.FromCurrentSynchronizationContext();

			InitializeComponent();

			PrintContextInfo("AppForm constructor");
		}

		private async void processTableButton_Click(object sender, EventArgs e) {
			eventLogListBox.Items.Clear();

			PrintContextInfo("processTableButton_Click starts");
			processTableButton.Enabled = false;

			using (var sr = new StreamReader(new FileStream("Zelenina.txt", FileMode.Open, FileAccess.Read, FileShare.Read, 16, true))) {
				Debug.WriteLine("Reader's stream: IsAsync == {0}", (sr.BaseStream as FileStream).IsAsync);
				using (var sw = new StreamWriter(new FileStream("Vystup.txt", FileMode.Create, FileAccess.Write, FileShare.Read, 16, true))) {
					Debug.WriteLine("Writer's stream: IsAsync == {0}", (sw.BaseStream as FileStream).IsAsync);

					if (((Button) sender).Tag.Equals("1")) {
						await Task.Factory.StartNew(() => {
							ProcessFile(sr, sw).Wait();
						});
					} else if (((Button) sender).Tag.Equals("2")) {
						//
					} else if (((Button) sender).Tag.Equals("3")) {
						await ProcessFile(sr, sw);
					} else if (((Button) sender).Tag.Equals("4")) {
						await ProcessFile(sr, sw).ConfigureAwait(continueOnCapturedContext: true);
					} else if (((Button) sender).Tag.Equals("5")) {
						await ProcessFile(sr, sw).ConfigureAwait(continueOnCapturedContext: false);
					}
				}
			}

			processTableButton.Enabled = true;
			PrintContextInfo("processTableButton_Click ends");
		}

		static readonly char[] Separators = { ' ', '\t' };

		static async Task ProcessFile(TextReader input, TextWriter output) {
			PrintContextInfo("● ProcessFile starts");

			var readLineTask = input.ReadLineAsync();
			Task writeLineTask = null;
			string line;
			while ((line = await readLineTask) != null) {
				readLineTask = input.ReadLineAsync();

				string[] words = line.Split(Separators, StringSplitOptions.RemoveEmptyEntries);

				PrintContextInfo("● ● ProcessFile loop iteration");
				await Task.Delay(100);

				if (writeLineTask != null) {
					await writeLineTask;
				}
				writeLineTask = output.WriteLineAsync(words[0]);
			}

			PrintContextInfo("● ProcessFile ends");
		}

		static void PrintContextInfo(string callerInfo = null) {
			int callingThreadId = System.Threading.Thread.CurrentThread.ManagedThreadId;
			string callingTaskId = Task.CurrentId == null ? "●null●" : Task.CurrentId.Value.ToString();
			string callingSynchroContext = System.Threading.SynchronizationContext.Current == null ? "●null●" : System.Threading.SynchronizationContext.Current.GetType().Name;

			new Task(() => {
				instance.eventLogListBox.Items.Add(
					string.Format("ThreadId: {0} ({4}), TaskId: {1} ({5}), SynchroContext: {2} ({6}) {3}",
						callingThreadId, callingTaskId, callingSynchroContext, callerInfo == null ? "" : "in " + callerInfo, 

						System.Threading.Thread.CurrentThread.ManagedThreadId,
						Task.CurrentId == null ? "●null●" : Task.CurrentId.Value.ToString(),
						System.Threading.SynchronizationContext.Current == null ? "●null●" : System.Threading.SynchronizationContext.Current.GetType().Name
					)
				);
			}).Start(uiTaskScheduler);
		}
	}
}
